<?php

/**
 * @file
 * Autolink text processors.
 */

/**
 * Define the cache interface.
 */
interface AutolinkPluginCache {
  public function setItems($id, $items);
  public function getItems($id);
  public function clearItems($id);
}

/**
 * Cache object class handles caching within processors. During preprocessing,
 * the cache object is passed to plugin callbacks, allowing them to easily
 * cache data in the Autolink cache.
 */
class AutolinkCache implements AutolinkPluginCache {

  // The machine-name of the current plugin. This is used in building
  // cache IDs for unique storage.
  var $plugin;

  // The current input format. This is also used for building cache IDs.
  var $format;

  public function __construct($plugin, $format) {
    $this->plugin = $plugin;
    $this->format = $format;
  }

  /**
   * Caches item arrays for specific plugins. This should be used by plugins
   * within the scope of the preprocess callback.
   */
  public function setItems($id, $items, $expire = CACHE_PERMANENT) {
    if (!empty($items) && is_array($items)) {
      cache_set($this->plugin . '_' . $this->format . '_items_' . $id, $items, AUTOLINK_CACHE, $expire);
    }
  }

  /**
   * Returns cached items from the Autolink cache. This may be used by plugins
   * within the scope of the preprocess callback.
   */
  public function getItems($id) {
    $items = cache_get($this->plugin . '_' . $this->format . '_items_' . $id, AUTOLINK_CACHE);
    return isset($items->data) && !empty($items->data) ? $items->data : FALSE;
  }

  /**
   * Clears the items cache for a specific plugin.
   */
  public function clearItems($id) {
    cache_clear_all($this->plugin . '_' . $this->format . '_items_' . $id, AUTOLINK_CACHE);
  }

}

/**
 * Defines the Autolink plugin interface.
 */
interface AutolinkPluginObject {
  public function init($plugin, $format, $plugin_info = array());
  public function getInfo();
  public function getCallback($name);
  public function save();
  public function reset();
}

/**
 * Plugins handle actual processing of text and generation of links.
 */
class AutolinkPlugin implements AutolinkPluginObject {

  // An AutolinkConfig object populated with settings for the current input format.
  protected $config;

  // The AutolinkCache object related to this plugin. Will handle caching data for the plugin.
  public $cache;

  // A string indicating the current input format.
  public $format;

  // An associative array of plugin information defined in hook_autolink_plugin_info().
  public $info = array();

  // The machine-name of the plugin.
  public $name = NULL;

  // The module that defines the plugin.
  public $module = NULL;

  // The human-readable name of the plugin.
  public $label = '';

  // The human-readable description of the plugin.
  public $description = '';

  // A string identifying the processor that is used by this plugin.
  public $processor = NULL;

  // An array of configuration settings as saved in plugin configuration pages
  // via the Drupal input formats interface.
  public $settings = array();

  public function __construct() {}

  /**
   * Initializes a plugin instance. Properties are populated from plugin info.
   *
   * @param $plugin
   *   The machine-name of the plugin to initialize.
   * @param $format
   *   The machine-name of the format related to this plugin instance.
   * @param $plugin_info
   *   An optional array of plugin info for the plugin being initialized.
   *   This parameter is primarily used for unit testing this method.
   * @return
   *   An initialized AutolinkPlugin object.
   */
  public function init($plugin, $format, $plugin_info = array()) {
    $config = autolink_config($format);
    $this->format = $format;
    if (!empty($plugin_info) || ($plugin_info = $config::$plugin_info[$plugin])) {
      $this->config = $config;
      $this->info = $plugin_info;
      $this->name = $plugin;
      $this->label = $plugin_info['label'];
      $this->description = isset($plugin_info['description']) ? $plugin_info['description'] : '';
      $this->processor = $plugin_info['processor'];
      $this->module = $plugin_info['module'];
      $this->cache = new AutolinkCache($plugin, $format);

      // If no settings have been saved yet then check for default settings.
      // It is possible for a plugin to define default settings from both within
      // plugin information and in a separate callback. The callback is called last.
      $this->settings = variable_get('autolink_plugin_' . $this->name . '_' . $this->format, array());
      if (empty($this->settings)) {
        $this->getDefaultSettings();
      }
    }
    return $this;
  }

  /**
   * Returns default configuration options for settings that are empty.
   * Configuration settings are retrieved from plugin info or callbacks.
   */
  protected function getDefaultSettings() {
    if (isset($this->info['default settings'])) {
      $this->settings = $this->info['default settings'];
    }
    if (isset($this->info['default settings callback']) && function_exists($this->info['default settings callback'])) {
      $defaults = call_user_func($this->info['default settings callback'], $this->name, $this->format);
      $this->settings = array_merge($this->settings, $defaults);
    }
  }

  /**
   * Magic setter method.
   */
  public function __set($property, $value) {
    $this->{$property} = $value;
  }

  /**
   * Magic call method for calling functions based on plugin info.
   *
   * This allows plugin modules to define their own callbacks and call
   * it by using $plugin->execute(Callback). We search for 'execute'
   * in the method and if found strip it and get the proper callback
   * from the plugin info. If no callback was identified then
   * {module}_{plugin}_{callback_name} is used.
   *
   * Since this magic callback is used many times during processing,
   * we statically cache functions that do not exist to prevent having
   * to repeat the check dozens of times in the same process.
   */
  public function __call($method, $args) {
    static $callbacks = array();
    // If the callback exists and is known it will be stored in the static array.
    if (isset($callbacks[$this->name][$method])) {
      // Callbacks stored in the static variable have already been checked.
      if (!empty($callbacks[$this->name][$method])) {
        return call_user_func_array($callbacks[$this->name][$method], $args);
      }
      // The callback was not found previously so return nothing.
      return;
    }
    // We haven't checked for this callback yet, so parse the method and check for it.
    elseif (strstr($method, 'execute')) {
      $name = drupal_strtolower(substr($method, 7));
      $callback = $this->getCallback($name);
      if (function_exists($callback)) {
        $callbacks[$this->name][$method] = $callback;
        return call_user_func_array($callback, $args);
      }
      // Callbacks that don't exist are noted in the static array as FALSE.
      else {
        $callbacks[$this->name][$method] = FALSE;
      }
    }
  }

  /**
   * Returns a callback string as identified in plugin information.
   *
   * Plugin modules may use this method to retrieve accurate callback information
   * defined in custom values in hook_autolink_plugin_info(). Any plugin can simply
   * add another key with the word 'callback' following the callback name and use
   * this method to retrieve that information. Also see the magic AutolinkPlugin::__call()
   * method for more use of this method.
   *
   * @param $name
   *   The prefix of the callback to retrieve as identified in plugin information.
   * @return
   *   The callback located in plugin information if it exists, otherwise, the method
   *   will return an alternative callback based on standard Autolink convention.
   *   This callback is formatted {MODULE}_{PLUGIN}_{CALLBACK}.
   * @see AutolinkPlugin::__call()
   */
  public function getCallback($name) {
    return isset($this->info[$name . ' callback']) ? $this->info[$name . ' callback'] : "{$this->module}_{$this->name}_{$name}";
  }

  /**
   * Returns plugin info.
   */
  public function getInfo() {
    return $this->info;
  }

  /**
   * Saves plugin settings.
   */
  public function save() {
    variable_set('autolink_plugin_' . $this->name . '_' . $this->format, $this->settings);
  }

  /**
   * Resets the plugin settings.
   */
  public function reset() {
    $this->settings = array();
    $this->getDefaultSettings();
  }

}
